/*
This is a port of a part of the difflib module in Python's standard
library to D (Python version 2.5).

The original file seems to be covered (any lawyers here?) by the
Python license (see file "Python_LICENSE.txt").


D port of SequenceMatcher class:
(C) 2006 Michael Butscher (mbutscher@gmx.de)


Module difflib -- helpers for computing deltas between objects.

Class SequenceMatcher:
    A flexible class for comparing pairs of sequences of any type.
*/


import std.stdio;
import std.string;


/*
// Helper for debugging
char[] toStringAA(K, V)(V[K] aa)
{
    char[] result;
    foreach(k; aa.keys)
    {
        result ~= format("%s: %s, ", k, aa[k]);
    }
    
    result ~= "\n";

    return result;
}
*/



double _calculate_ratio(long matches, long length)
{
    if (length)
        return (2.0 * matches) / length;
    return 1.0;
}


/**
SequenceMatcher is a flexible class for comparing pairs of sequences of
any type, so long as the sequence elements are hashable.  The basic
algorithm predates, and is a little fancier than, an algorithm
published in the late 1980's by Ratcliff and Obershelp under the
hyperbolic name "gestalt pattern matching".  The basic idea is to find
the longest contiguous matching subsequence that contains no "junk"
elements (R-O doesn't address junk).  The same idea is then applied
recursively to the pieces of the sequences to the left and to the right
of the matching subsequence.  This does not yield minimal edit
sequences, but does tend to yield matches that "look right" to people.

SequenceMatcher tries to compute a "human-friendly diff" between two
sequences.  Unlike e.g. UNIX(tm) diff, the fundamental notion is the
longest *contiguous* & junk-free matching subsequence.  That's what
catches peoples' eyes.  The Windows(tm) windiff has another interesting
notion, pairing up elements that appear uniquely in each sequence.
That, and the method here, appear to yield more intuitive difference
reports than does diff.  This method appears to be the least vulnerable
to synching up on blocks of "junk lines", though (like blank lines in
ordinary text files, or maybe "<P>" lines in HTML files).  That may be
because this is the only method of the 3 that has a *concept* of
"junk" <wink>.

Example, comparing two strings, and considering blanks to be "junk":

>>> s = SequenceMatcher(lambda x: x == " ",
...                     "private Thread currentThread;",
...                     "private volatile Thread currentThread;")
>>>

.ratio() returns a float in [0, 1], measuring the "similarity" of the
sequences.  As a rule of thumb, a .ratio() value over 0.6 means the
sequences are close matches:

>>> print round(s.ratio(), 3)
0.866
>>>

If you're only interested in where the sequences match,
.get_matching_blocks() is handy:

>>> for block in s.get_matching_blocks():
...     print "a[%d] and b[%d] match for %d elements" % block
a[0] and b[0] match for 8 elements
a[8] and b[17] match for 21 elements
a[29] and b[38] match for 0 elements

Note that the last tuple returned by .get_matching_blocks() is always a
dummy, (len(a), len(b), 0), and this is the only case in which the last
tuple element (number of elements matched) is 0.

If you want to know how to change the first sequence into the second,
use .get_opcodes():

>>> for opcode in s.get_opcodes():
...     print "%6s a[%d:%d] b[%d:%d]" % opcode
 equal a[0:8] b[0:8]
insert a[8:8] b[8:17]
 equal a[8:29] b[17:38]

See the Differ class for a fancy human-friendly file differencer, which
uses SequenceMatcher both to compare sequences of lines, and to compare
sequences of characters within similar (near-matching) lines.

See also function get_close_matches() in this module, which shows how
simple code building on SequenceMatcher can be used to do useful work.

Timing:  Basic R-O is cubic time worst case and quadratic time expected
case.  SequenceMatcher is quadratic time for the worst case and has
expected-case behavior dependent in a complicated way on how many
elements the sequences have in common; best case time is linear.

Methods:

__init__(isjunk=None, a='', b='')
    Construct a SequenceMatcher.

set_seqs(a, b)
    Set the two sequences to be compared.

set_seq1(a)
    Set the first sequence to be compared.

set_seq2(b)
    Set the second sequence to be compared.

find_longest_match(alo, ahi, blo, bhi)
    Find longest matching block in a[alo:ahi] and b[blo:bhi].

get_matching_blocks()
    Return list of triples describing matching subsequences.

get_opcodes()
    Return list of 5-tuples describing how to turn a into b.

ratio()
    Return a measure of the sequences' similarity (float in [0,1]).

quick_ratio()
    Return an upper bound on .ratio() relatively quickly.

real_quick_ratio()
    Return an upper bound on ratio() very quickly.
*/

class SequenceMatcher(ET, ST=ET[], IT=int)
{
    // ET: Type of sequence element
    // ST: Type of sequence (must behave like an array)
    // IT: Type of index into sequence (must be signed)


    public:
    
    // Types
    
    alias bool delegate(ET elem) JunkDeg;
    alias ET ElementType;
    alias ST SequenceType;
    alias IT IndexType;  /// Must be signed

    struct Match
    {
        IndexType i;
        IndexType j;
        IndexType len;
        
        static Match create(IndexType _i, IndexType _j, IndexType _len)
        {
            Match result;
            result.i = _i;
            result.j = _j;
            result.len = _len;
            
            return result;
        }
        
        int opCmp(Match* pOther)
        {
            int result = i - (*pOther).i;
            if (result != 0)
                return result;
            
            result = j - (*pOther).j;
            if (result != 0)
                return result;
            
            return len - (*pOther).len;
        }
        
        int opEquals(Match other)
        {
            return opCmp(&other) == 0;
        }
        
        char[] toString()
        {
            return "difflib.Match("~.toString(i)~", "~.toString(j)~", "~
                    .toString(len)~")";
        }

    }


    /**
    a struct containing (tag, i1, i2, j1, j2) where tag is
    one of
        'replace'   a[i1:i2] should be replaced by b[j1:j2]
        'delete'    a[i1:i2] should be deleted
        'insert'    b[j1:j2] should be inserted
        'equal'     a[i1:i2] == b[j1:j2]
    */
    struct Opcode
    {
        const char[] REPLACE = "replace";
        const char[] DELETE = "delete";
        const char[] INSERT = "insert";
        const char[] EQUAL = "equal";

        char[] tag;
        IndexType i1, i2;
        IndexType j1, j2;
        
        static Opcode create(char[] _tag, IndexType _i1, IndexType _i2,
                IndexType _j1, IndexType _j2)
        {
            Opcode result;
            result.tag = _tag;
            result.i1 = _i1;
            result.i2 = _i2;
            result.j1 = _j1;
            result.j2 = _j2;
            
            return result;
        }
        
        char[] toString()
        {
            return "difflib.Opcode(\""~tag~"\", "~.toString(i1)~", "~.toString(i2)~
                    ", "~.toString(j1)~", "~.toString(j2)~")";
        }
        
        int opEquals(Opcode other)
        {
            return tag == other.tag && i1 == other.i1 && i2 == other.i2 &&
                    j1 == other.j1 && j2 == other.j2;
        }
    }
 

    protected:
    
    ST a;   /// First sequence
    ST b;   /// Second sequence
    
    /**
    a user-supplied function taking a sequence element and
    returning true iff the element is "junk" -- this has
    subtle but helpful effects on the algorithm, which I'll
    get around to writing up someday <0.9 wink>.
    DON'T USE!  Only chain_b uses this.  Use isbjunk.
    */ 
    JunkDeg isjunk;

    IndexType[ET] populardict;
    IndexType[ET] junkdict;


    /**
    for x in b, b2j[x] is a list of the indices (into b)
    at which x appears; junk elements do not appear
    */
    IndexType[][ET] b2j;  


    /**
          a list of (i, j, k) triples, where a[i:i+k] == b[j:j+k];
          ascending & non-overlapping in i and in j; terminated by
          a dummy (len(a), len(b), 0) sentinel
    */
    Match[] matching_blocks;


    /**
          an array of Opcode(tag, i1, i2, j1, j2) structs, where tag is
          one of
              'replace'   a[i1:i2] should be replaced by b[j1:j2]
              'delete'    a[i1:i2] should be deleted
              'insert'    b[j1:j2] should be inserted
              'equal'     a[i1:i2] == b[j1:j2]
    */
    Opcode[] opcodes;

    
    IndexType[ET] fullbcount;


    public:

    /** Construct a SequenceMatcher.

    Optional arg isjunk is None (the default), or a one-argument
    function that takes a sequence element and returns true iff the
    element is junk.  None is equivalent to passing "lambda x: 0", i.e.
    no elements are considered to be junk.  For example, pass
        lambda x: x in " \\t"
    if you're comparing lines as sequences of characters, and don't
    want to synch up on blanks or hard tabs.

    Optional arg a is the first of two sequences to be compared.  By
    default, an empty string.  The elements of a must be hashable.  See
    also .set_seqs() and .set_seq1().

    Optional arg b is the second of two sequences to be compared.  By
    default, an empty string.  The elements of b must be hashable. See
    also .set_seqs() and .set_seq2().
    */
    this(JunkDeg isjunk=null, ST a=null, ST b=null)
    {
        // Members:
        // a
        //      first sequence
        // b
        //      second sequence; differences are computed as "what do
        //      we need to do to 'a' to change it into 'b'?"
        // b2j
        //      for x in b, b2j[x] is a list of the indices (into b)
        //      at which x appears; junk elements do not appear
        // fullbcount
        //      for x in b, fullbcount[x] == the number of times x
        //      appears in b; only materialized if really needed (used
        //      only for computing quick_ratio())
        // matching_blocks
        //      a list of (i, j, k) triples, where a[i:i+k] == b[j:j+k];
        //      ascending & non-overlapping in i and in j; terminated by
        //      a dummy (len(a), len(b), 0) sentinel
        // opcodes
        //      a list of (tag, i1, i2, j1, j2) tuples, where tag is
        //      one of
        //          'replace'   a[i1:i2] should be replaced by b[j1:j2]
        //          'delete'    a[i1:i2] should be deleted
        //          'insert'    b[j1:j2] should be inserted
        //          'equal'     a[i1:i2] == b[j1:j2]
        // isjunk
        //      a user-supplied function taking a sequence element and
        //      returning true iff the element is "junk" -- this has
        //      subtle but helpful effects on the algorithm, which I'll
        //      get around to writing up someday <0.9 wink>.
        //      DON'T USE!  Only chain_b uses this.  Use isbjunk.
        // isbjunk
        //      for x in b, isbjunk(x) == isjunk(x) but much faster;
        //      it's really the has_key method of a hidden dict.
        //      DOES NOT WORK for x in a!
        // isbpopular
        //      for x in b, isbpopular(x) is true iff b is reasonably long
        //      (at least 200 elements) and x accounts for more than 1% of
        //      its elements.  DOES NOT WORK for x in a!

        this.isjunk = isjunk;
        this.a = this.b = null;
        set_seqs(a, b);
    }

    /**
    Set the two sequences to be compared.

    >>> s = SequenceMatcher()
    >>> s.set_seqs("abcd", "bcde")
    >>> s.ratio()
    0.75
    */
    void set_seqs(ST a, ST b)
    {
        set_seq1(a);
        set_seq2(b);
    }
    
    /**
    Set the first sequence to be compared.

    The second sequence to be compared is not changed.

    >>> s = SequenceMatcher(None, "abcd", "bcde")
    >>> s.ratio()
    0.75
    >>> s.set_seq1("bcde")
    >>> s.ratio()
    1.0
    >>>

    SequenceMatcher computes and caches detailed information about the
    second sequence, so if you want to compare one sequence S against
    many sequences, use .set_seq2(S) once and call .set_seq1(x)
    repeatedly for each of the other sequences.

    See also set_seqs() and set_seq2().
    */
    void set_seq1(ST a)
    {
        if (a is this.a)
            return;
        this.a = a;
        this.matching_blocks = null;
        this.opcodes = null;
    }

    /**
    Set the second sequence to be compared.

    The first sequence to be compared is not changed.

    >>> s = SequenceMatcher(None, "abcd", "bcde")
    >>> s.ratio()
    0.75
    >>> s.set_seq2("abcd")
    >>> s.ratio()
    1.0
    >>>

    SequenceMatcher computes and caches detailed information about the
    second sequence, so if you want to compare one sequence S against
    many sequences, use .set_seq2(S) once and call .set_seq1(x)
    repeatedly for each of the other sequences.

    See also set_seqs() and set_seq1().
    */
    void set_seq2(ST b)
    {
        if (b is this.b)
            return;
        this.b = b;
        this.matching_blocks = null;
        this.opcodes = null;
        this.fullbcount = null;
        this.chain_b();
    }

    /**
    For each element x in b, set b2j[x] to a list of the indices in
    b where x appears; the indices are in increasing order; note that
    the number of times x appears in b is len(b2j[x]) ...
    when self.isjunk is defined, junk elements don't show up in this
    map at all, which stops the central find_longest_match method
    from starting any matching block at a junk element ...
    also creates the fast isbjunk function ...
    b2j also does not contain entries for "popular" elements, meaning
    elements that account for more than 1% of the total elements, and
    when the sequence is reasonably large (>= 200 elements); this can
    be viewed as an adaptive notion of semi-junk, and yields an enormous
    speedup when, e.g., comparing program files with hundreds of
    instances of "return NULL;" ...
    note that this is only called when b changes; so for cross-product
    kinds of matches, it's best to call set_seq2 once, then set_seq1
    repeatedly
    */
    private void chain_b()
    {
        // Because isjunk is a user-defined (not C) function, and we test
        // for junk a LOT, it's important to minimize the number of calls.
        // Before the tricks described here, chain_b was by far the most
        // time-consuming routine in the whole module!  If anyone sees
        // Jim Roskind, thank him again for profile.py -- I never would
        // have guessed that.
        // The first trick is to build b2j ignoring the possibility
        // of junk.  I.e., we don't call isjunk at all yet.  Throwing
        // out the junk later is much cheaper than building b2j "right"
        // from the start.
        ST b = this.b;
        IndexType n = b.length;
        IndexType[][ET] b2j;
        this.b2j = null;
        
        IndexType[ET] populardict;
        IndexType i = 0;

        foreach(ET elt; b)   // TODO Does not allow general sequence types
        {
            IndexType[]* pIndices = elt in b2j;
            
            if (pIndices)
            {
                if ((n >= 200) && ((*pIndices).length * 100 > n))
                {
                    populardict[elt] = 1;
                    (*pIndices) = null;
                }
                else
                    (*pIndices) ~= i;
            }
            else
                b2j[elt] = [i];
            
            ++i;
        }

        // Purge leftover indices for popular elements.
        foreach(ET elt; populardict.keys)
            b2j.remove(elt);

        // Now b2j.keys() contains elements uniquely, and especially when
        // the sequence is a string, that's usually a good deal smaller
        // than len(string).  The difference is the number of isjunk calls
        // saved.
        JunkDeg isjunk = this.isjunk;
        IndexType[ET] junkdict;
        if (isjunk)
        {
            foreach(ET elt; populardict.keys)
            {
                if (isjunk(elt))
                {
                    junkdict[elt] = 1;
                    populardict.remove(elt);
                }
            }
            foreach(ET elt; b2j.keys)
            {
                if (isjunk(elt))
                {
                    junkdict[elt] = 1;
                    b2j.remove(elt);
                }
            }
        }

        // Now for x in b, isjunk(x) == x in junkdict, but the
        // latter is much faster.  Note too that while there may be a
        // lot of junk in the sequence, the number of *unique* junk
        // elements is probably small.  So the memory burden of keeping
        // this dict alive is likely trivial compared to the size of b2j.
        this.junkdict = junkdict;
        this.populardict = populardict;
        this.b2j = b2j;
    }
    
    bool isbjunk(ET elt)
    {
        return (elt in this.junkdict) !is null;
    }
    
    bool isbpopular(ET elt)
    {
        return (elt in this.populardict) !is null;
    }
    


    /**
    Find longest matching block in a[alo:ahi] and b[blo:bhi].

    If isjunk is not defined:

    Return (i,j,k) such that a[i:i+k] is equal to b[j:j+k], where
        alo <= i <= i+k <= ahi
        blo <= j <= j+k <= bhi
    and for all (i',j',k') meeting those conditions,
        k >= k'
        i <= i'
        and if i == i', j <= j'

    In other words, of all maximal matching blocks, return one that
    starts earliest in a, and of all those maximal matching blocks that
    start earliest in a, return the one that starts earliest in b.

    >>> s = SequenceMatcher(None, " abcd", "abcd abcd")
    >>> s.find_longest_match(0, 5, 0, 9)
    (0, 4, 5)

    If isjunk is defined, first the longest matching block is
    determined as above, but with the additional restriction that no
    junk element appears in the block.  Then that block is extended as
    far as possible by matching (only) junk elements on both sides.  So
    the resulting block never matches on junk except as identical junk
    happens to be adjacent to an "interesting" match.

    Here's the same example as before, but considering blanks to be
    junk.  That prevents " abcd" from matching the " abcd" at the tail
    end of the second sequence directly.  Instead only the "abcd" can
    match, and matches the leftmost "abcd" in the second sequence:

    >>> s = SequenceMatcher(lambda x: x==" ", " abcd", "abcd abcd")
    >>> s.find_longest_match(0, 5, 0, 9)
    (1, 0, 4)

    If no blocks match, return (alo, blo, 0).

    >>> s = SequenceMatcher(None, "ab", "c")
    >>> s.find_longest_match(0, 2, 0, 1)
    (0, 0, 0)
    */
    
    Match find_longest_match(IndexType alo, IndexType ahi, IndexType blo,
            IndexType bhi)
    {
        // CAUTION:  stripping common prefix or suffix would be incorrect.
        // E.g.,
        //    ab
        //    acab
        // Longest matching block is "ab", but if common prefix is
        // stripped, it's "a" (tied with "b").  UNIX(tm) diff does so
        // strip, so ends up claiming that ab is changed to acab by
        // inserting "ca" in the middle.  That's minimal but unintuitive:
        // "it's obvious" that someone inserted "ac" at the front.
        // Windiff ends up at the same place as diff, but by pairing up
        // the unique 'b's and then matching the first two 'a's.
        
        ST a = this.a;
        ST b = this.b;
        IndexType[][ET] b2j = this.b2j;
        
        IndexType besti = alo;
        IndexType bestj = blo;
        IndexType bestsize = 0;
        
        // find longest junk-free match
        // during an iteration of the loop, j2len[j] = length of longest
        // junk-free match ending with a[i-1] and b[j]
        IndexType[IndexType] j2len;
        const IndexType[] nothing = null;
        
        IndexType j2lenget(IndexType i, IndexType def)
        {
            IndexType* result = i in j2len;
            if (result)
                return *result;
            else
                return def;
        }

        IndexType[] b2jget(ET i, IndexType[] def)
        {
            IndexType[]* result = i in b2j;
            if (result)
                return *result;
            else
                return def;
        }
            
        for(IndexType i = alo; i < ahi; ++i)
        {
            // look at all instances of a[i] in b; note that because
            // b2j has no junk keys, the loop is skipped if a[i] is junk
            IndexType[IndexType] newj2len;
            IndexType k;
            
            foreach(IndexType j; b2jget(a[i], nothing))
            {
                if (j < blo)
                    continue;
                if (j >= bhi)
                    break;
                k = newj2len[j] = j2lenget(j-1, 0) + 1;
                if (k > bestsize)
                {
                    besti = i-k+1;
                    bestj = j-k+1;
                    bestsize = k;
                }
            }
            j2len = newj2len;
        }
        
        // Extend the best by non-junk elements on each end.  In particular,
        // "popular" non-junk elements aren't in b2j, which greatly speeds
        // the inner loop above, but also means "the best" match so far
        // doesn't contain any junk *or* popular non-junk elements.
        
        while ( besti > alo && bestj > blo && 
              !isbjunk(b[bestj-1]) && 
              a[besti-1] == b[bestj-1] )
        {
            besti = besti-1;
            bestj = bestj-1;
            bestsize = bestsize+1;
        }

        while ( besti+bestsize < ahi && bestj+bestsize < bhi &&
              !isbjunk(b[bestj+bestsize]) &&
              a[besti+bestsize] == b[bestj+bestsize] )
        {
            ++bestsize;
        }

        // Now that we have a wholly interesting match (albeit possibly
        // empty!), we may as well suck up the matching junk on each
        // side of it too.  Can't think of a good reason not to, and it
        // saves post-processing the (possibly considerable) expense of
        // figuring out what to do with it.  In the case of an empty
        // interesting match, this is clearly the right thing to do,
        // because no other kind of match is possible in the regions.
        while ( besti > alo && bestj > blo &&
              isbjunk(b[bestj-1]) &&
              a[besti-1] == b[bestj-1] )
        {
            besti = besti-1;
            bestj = bestj-1;
            bestsize = bestsize+1;
        }

        while ( besti+bestsize < ahi && bestj+bestsize < bhi &&
              isbjunk(b[bestj+bestsize]) &&
              a[besti+bestsize] == b[bestj+bestsize] )
        {
            ++bestsize;
        }
        
        Match result;
        result.i = besti;
        result.j = bestj;
        result.len = bestsize;

        return result;
    }


    /**
    Return list of triples describing matching subsequences.

    Each triple is of the form (i, j, n), and means that
    a[i:i+n] == b[j:j+n].  The triples are monotonically increasing in
    i and in j.  New in Python 2.5, it's also guaranteed that if
    (i, j, n) and (i', j', n') are adjacent triples in the list, and
    the second is not the last triple in the list, then i+n != i' or
    j+n != j'.  IOW, adjacent triples never describe adjacent equal
    blocks.

    The last triple is a dummy, (len(a), len(b), 0), and is the only
    triple with n==0.

    >>> s = SequenceMatcher(None, "abxcd", "abcd")
    >>> s.get_matching_blocks()
    [(0, 0, 2), (3, 2, 2), (5, 4, 0)]
    */
    
    Match[] get_matching_blocks()
    {
        struct QueueEntry
        {
            IndexType a1, a2;
            IndexType b1, b2;
            
            static QueueEntry create(IndexType _a1, IndexType _a2, IndexType _b1,
                    IndexType _b2)
            {
                QueueEntry result;
                result.a1 = _a1;
                result.a2 = _a2;
                result.b1 = _b1;
                result.b2 = _b2;
                
                return result;
            }
        }
                
            
        if (this.matching_blocks !is null)
            return this.matching_blocks;

        IndexType la = this.a.length;
        IndexType lb = this.b.length;

        // This is most naturally expressed as a recursive algorithm, but
        // at least one user bumped into extreme use cases that exceeded
        // the recursion limit on their box.  So, now we maintain a list
        // ('queue`) of blocks we still need to look at, and append partial
        // results to `matching_blocks` in a loop; the matches are sorted
        // at the end.

        QueueEntry[] queue = [QueueEntry.create(0, la, 0, lb)];
        Match[] matching_blocks;
        while (queue.length > 0)
        {
            IndexType alo, ahi, blo, bhi;
            with(queue[length-1])
            {
                alo = a1;
                ahi = a2;
                blo = b1;
                bhi = b2;
            }
            queue.length = queue.length - 1;
            
            Match x = find_longest_match(alo, ahi, blo, bhi);
            // a[alo:i] vs b[blo:j] unknown
            // a[i:i+k] same as b[j:j+k]
            // a[i+k:ahi] vs b[j+k:bhi] unknown
            if (x.len > 0)   // if k is 0, there was no matching block
            {
                matching_blocks ~= x;
                if (alo < x.i && blo < x.j)
                    queue ~= QueueEntry.create(alo, x.i, blo, x.j);
                if (x.i+x.len < ahi && x.j+x.len < bhi)
                    queue ~= QueueEntry.create(x.i+x.len, ahi, x.j+x.len, bhi);
            }
        }
        matching_blocks.sort;

        // It's possible that we have adjacent equal blocks in the
        // matching_blocks list now.  Starting with 2.5, this code was added
        // to collapse them.
        IndexType i1 = 0, j1 = 0, k1 = 0;
        Match[] non_adjacent;
        foreach(Match m; matching_blocks)
        {
            // Is this block adjacent to i1, j1, k1?
            if ((i1 + k1 == m.i) && (j1 + k1 == m.j))
            {
                // Yes, so collapse them -- this just increases the length of
                // the first block by the length of the second, and the first
                // block so lengthened remains the block to compare against.
                k1 += m.len;
            }
            else
            {
                // Not adjacent.  Remember the first block (k1==0 means it's
                // the dummy we started with), and make the second block the
                // new block to compare against.
                if (k1 > 0)
                    non_adjacent ~= Match.create(i1, j1, k1);

                i1 = m.i;
                j1 = m.j;
                k1 = m.len;
            }
        }
        if (k1 > 0)
            non_adjacent ~= Match.create(i1, j1, k1);

        non_adjacent ~= Match.create(la, lb, 0);
        this.matching_blocks = non_adjacent;

        return this.matching_blocks;
    }




    /**
    Return list of 5-tuples describing how to turn a into b.

    Each tuple is of the form (tag, i1, i2, j1, j2).  The first tuple
    has i1 == j1 == 0, and remaining tuples have i1 == the i2 from the
    tuple preceding it, and likewise for j1 == the previous j2.

    The tags are strings, with these meanings:

    'replace':  a[i1:i2] should be replaced by b[j1:j2]
    'delete':   a[i1:i2] should be deleted.
                Note that j1==j2 in this case.
    'insert':   b[j1:j2] should be inserted at a[i1:i1].
                Note that i1==i2 in this case.
    'equal':    a[i1:i2] == b[j1:j2]

    >>> a = "qabxcd"
    >>> b = "abycdf"
    >>> s = SequenceMatcher(None, a, b)
    >>> for tag, i1, i2, j1, j2 in s.get_opcodes():
    ...    print ("%7s a[%d:%d] (%s) b[%d:%d] (%s)" %
    ...           (tag, i1, i2, a[i1:i2], j1, j2, b[j1:j2]))
     delete a[0:1] (q) b[0:0] ()
      equal a[1:3] (ab) b[0:2] (ab)
    replace a[3:4] (x) b[2:3] (y)
      equal a[4:6] (cd) b[3:5] (cd)
     insert a[6:6] () b[5:6] (f)
    */

    Opcode[] get_opcodes()
    {
        if (this.opcodes !is null)
            return this.opcodes;
            
        IndexType i = 0, j = 0;
        
        this.opcodes = null;
        Opcode[] answer = null;
        
        foreach(Match m; get_matching_blocks())
        {
            // invariant:  we've pumped out correct diffs to change
            // a[:i] into b[:j], and the next matching block is
            // a[m.i:m.i+size] == b[m.j:m.j+size].  So we need to pump
            // out a diff to change a[i:m.i] into b[j:bj], pump out
            // the matching block, and move (i,j) beyond the match
            char[] tag;
            if (i < m.i && j < m.j)
                tag = Opcode.REPLACE;
            else if (i < m.i)
                tag = Opcode.DELETE;
            else if (j < m.j)
                tag = Opcode.INSERT;
            if (tag)
                answer ~= Opcode.create(tag, i, m.i, j, m.j);
            i = m.i + m.len;
            j = m.j + m.len;

            // the list of matching blocks is terminated by a
            // sentinel with size 0

            if (m.len)
                answer ~= Opcode.create(Opcode.EQUAL, m.i, i, m.j, j);
        }

        this.opcodes = answer;        

        return answer;
    }

    /**
    Isolate change clusters by eliminating ranges with no changes.

    Return an array of groups with upto n lines of context.
    Each group is in the same format as returned by get_opcodes().

    >>> from pprint import pprint
    >>> a = map(str, range(1,40))
    >>> b = a[:]
    >>> b[8:8] = ['i']     // Make an insertion
    >>> b[20] += 'x'       // Make a replacement
    >>> b[23:28] = []      // Make a deletion
    >>> b[30] += 'y'       // Make another replacement
    >>> pprint(list(SequenceMatcher(None,a,b).get_grouped_opcodes()))
    [[('equal', 5, 8, 5, 8), ('insert', 8, 8, 8, 9), ('equal', 8, 11, 9, 12)],
     [('equal', 16, 19, 17, 20),
      ('replace', 19, 20, 20, 21),
      ('equal', 20, 22, 21, 23),
      ('delete', 22, 27, 23, 23),
      ('equal', 27, 30, 23, 26)],
     [('equal', 31, 34, 27, 30),
      ('replace', 34, 35, 30, 31),
      ('equal', 35, 38, 31, 34)]]
    */

    Opcode[][] get_grouped_opcodes(IndexType n=3)
    {
        static IndexType min(IndexType a, IndexType b)
        {
            return (a < b) ? a : b;
        }

        static IndexType max(IndexType a, IndexType b)
        {
            return (a < b) ? b : a;
        }

        Opcode[] codes = get_opcodes().dup;
        if (!codes)
            codes = [Opcode.create(Opcode.EQUAL, 0, 1, 0, 1)];

        // Fixup leading and trailing groups if they show no changes.
        if (codes[0].tag == Opcode.EQUAL)
        {
            codes[0] = Opcode.create(codes[0].tag,
                    max(codes[0].i1, codes[0].i2 - n), codes[0].i2,
                    max(codes[0].j1, codes[0].j2 - n), codes[0].j2);
        }
        if (codes[length - 1].tag == Opcode.EQUAL)
        {
            IndexType top = codes.length - 1;
            codes[top] = Opcode.create(codes[top].tag,
                    codes[top].i1, min(codes[top].i2, codes[top].i1 + n),
                    codes[top].j1, min(codes[top].j2, codes[top].j1 + n));
        }

        IndexType nn = n + n;
        Opcode[] group;
        Opcode[][] result;
        
        foreach(opcode; codes)
        {
            // End the current group and start a new one whenever
            // there is a large range with no changes.
            if ((opcode.tag == Opcode.EQUAL) && opcode.i2 - opcode.i1 > nn)
            {
                group ~= Opcode.create(opcode.tag,
                        opcode.i1, min(opcode.i2, opcode.i1 + n),
                        opcode.j1, min(opcode.j2, opcode.j1 + n));
                result ~= group;
                group = null;
                opcode.i1 = max(opcode.i1, opcode.i2 - n);
                opcode.j1 = max(opcode.j1, opcode.j2 - n);
            }
            group ~= Opcode.create(opcode.tag, opcode.i1, opcode.i2, opcode.j1,
                    opcode.j2);
        }
        if (group && !(group.length == 1 && group[0].tag == Opcode.EQUAL))
            result ~= group;

        return result;
    }
 
    

    /**
    Return a measure of the sequences' similarity (float in [0,1]).

    Where T is the total number of elements in both sequences, and
    M is the number of matches, this is 2.0*M / T.
    Note that this is 1 if the sequences are identical, and 0 if
    they have nothing in common.

    .ratio() is expensive to compute if you haven't already computed
    .get_matching_blocks() or .get_opcodes(), in which case you may
    want to try .quick_ratio() or .real_quick_ratio() first to get an
    upper bound.

    >>> s = SequenceMatcher(None, "abcd", "bcde")
    >>> s.ratio()
    0.75
    >>> s.quick_ratio()
    0.75
    >>> s.real_quick_ratio()
    1.0
    */
    double ratio()
    {
        IndexType matches = 0;
        foreach(Match m; get_matching_blocks())
            matches += m.len;

        return _calculate_ratio(matches, this.a.length + this.b.length);
    }


    /**
    Return an upper bound on ratio() relatively quickly.

    This isn't defined beyond that it is an upper bound on .ratio(), and
    is faster to compute.
    */
    double quick_ratio()
    {
        // viewing a and b as multisets, set matches to the cardinality
        // of their intersection; this counts the number of matches
        // without regard to order, so is clearly an upper bound

        IndexType[ET] fullbcount;
        if (this.fullbcount is null)
        {
            foreach(elt; this.b)
            {
                IndexType* ct = elt in fullbcount;
                if (ct)
                    ++(*ct);
                else
                    fullbcount[elt] = 1;
            }
        }
        else
            fullbcount = this.fullbcount;

        // avail[x] is the number of times x appears in 'b' less the
        // number of times we've seen it in 'a' so far ... kinda
        IndexType[ET] avail;
        IndexType matches = 0;
        IndexType numb;

        foreach(elt; this.a)
        {
            IndexType* avel = elt in avail;
            if (avel)
                numb = *avel;
            else
                numb = (elt in fullbcount) ? fullbcount[elt] : 0;

            avail[elt] = numb - 1;
            if (numb > 0)
                ++matches;
        }
                
        this.fullbcount = fullbcount;
        return _calculate_ratio(matches, this.a.length + this.b.length);
    }


    /**
    Return an upper bound on ratio() very quickly.

    This isn't defined beyond that it is an upper bound on .ratio(), and
    is faster to compute than either .ratio() or .quick_ratio().
    */
    double real_quick_ratio()
    {
        IndexType la = this.a.length, lb = this.b.length;
        IndexType min = (la < lb) ? la : lb;
        // can't have more matches than the number of elements in the
        // shorter sequence
        return _calculate_ratio(min, la + lb);
    }
}



unittest
{
    alias SequenceMatcher!(char).Match CharMatch;

    auto s = new SequenceMatcher!(char)(delegate bool(char x) {return x == ' ';},
             " abcd", "abcd abcd");
    assert(s.find_longest_match(0, 5, 0, 9) == CharMatch.create(1, 0, 4));

    s = new SequenceMatcher!(char)(null, "ab", "c");
    
    assert(s.find_longest_match(0, 2, 0, 1) == CharMatch.create(0, 0, 0));
}

unittest
{
    alias SequenceMatcher!(char).Match CharMatch;

    auto s = new SequenceMatcher!(char)(null, "abxcd", "abcd");
    
    assert(s.get_matching_blocks() == [CharMatch.create(0, 0, 2),
            CharMatch.create(3, 2, 2), CharMatch.create(5, 4, 0)]);
}

unittest
{
    alias SequenceMatcher!(char).Opcode CharOpcode;    
    char[] a = "qabxcd";
    char[] b = "abycdf";
    auto s = new SequenceMatcher!(char)(null, a, b);

    assert(s.get_opcodes() == [
            CharOpcode.create("delete", 0, 1, 0, 0),
            CharOpcode.create("equal", 1, 3, 0, 2),
            CharOpcode.create("replace", 3, 4, 2, 3),
            CharOpcode.create("equal", 4, 6, 3, 5),
            CharOpcode.create("insert", 6, 6, 5, 6)
            ]);
}

/*
unittest
{
    alias SequenceMatcher!(char).Opcode CharOpcode;    

    char[][] a = [cast(char[])"1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11",
            "12", "13", "14", "15", "16", "17", "18", "19", "20", "21", "22",
            "23", "24", "25", "26", "27", "28", "29", "30", "31", "32", "33",
            "34", "35", "36", "37", "38", "39"];
    char[][] b = [cast(char[])"1", "2", "3", "4", "5", "6", "7", "8", "i", "9", "10", "11",
            "12", "13", "14", "15", "16", "17", "18", "19", "20x", "21", "22",
            "28","29", "30", "31", "32", "33", "34", "35y", "36", "37", "38",
            "39"];

    auto s = new SequenceMatcher!(char[])(null, a, b);

    writefln("test get_grouped_opcodes7 %s", s.get_grouped_opcodes());

    assert(s.get_grouped_opcodes() == cast(CharOpcode[][])[
            cast(CharOpcode[])[
                CharOpcode.create("equal", 5, 8, 5, 8),
                CharOpcode.create("insert", 8, 8, 8, 9),
                CharOpcode.create("equal", 8, 11, 9, 12)
            ],
            cast(CharOpcode[])[
                CharOpcode.create("equal", 16, 19, 17, 20),
                CharOpcode.create("replace", 19, 20, 20, 21),
                CharOpcode.create("equal", 20, 22, 21, 23),
                CharOpcode.create("delete", 22, 27, 23, 23),
                CharOpcode.create("equal", 27, 30, 23, 26)
            ],
            cast(CharOpcode[])[
                CharOpcode.create("equal", 31, 34, 27, 30),
                CharOpcode.create("replace", 34, 35, 30, 31),
                CharOpcode.create("equal", 35, 38, 31, 34)
            ]]);
}
*/


unittest
{
    auto s = new SequenceMatcher!(char)();

    s.set_seqs("abcd", "bcde");
    assert(s.ratio() == 0.75);
    assert(s.quick_ratio() == 0.75);
    assert(s.real_quick_ratio() == 1.0);
}

